//  Copyright (C) 2015 Florent Bedoiseau (buzz.computer@free.fr)
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
// Creation date : 31/08/2015

//
// Simple maze game
//
// As any maze game, you start from somewhere (red square) and you have to find 
// the exit (green square) as fast as possible.
// Each step is counted, so at the end, you will know how many steps you have walked
// in conparaison to the minimal needed steps to escape.

// Keys :
// R : Restart the game. A new maze is generated and time is cleared
// SPACE : Start the game
// Q : Left
// S : Right
// Z : Up
// W : Down
color white = color (0xFF, 0xFF, 0xFF);
color blue = color (0x00, 0x00, 0xFF);
color red = color (0xFF, 0x00, 0x00);
color green = color (0x00, 0xFF, 0x00);
color black = color (0x00, 0x00, 0x00);

int state_init = 0;
int state_run = 1;
int state_end = 2;

//=============== GAME ================
class Game {
  Game () {
    Reset();
  }
  
  void Reset () {
    _state = state_init;
     
    _maze = new Maze (20, 40, int (random(1, 6)));
    _maze.compute ();
    _maze.show (8);
    
    _needToRedraw = true;
  
    _startTime = 0;
    _endTime = 0;

    ClearTextArea();
  
    textAlign(CENTER);
    fill(black);
    text("Press SPACE to start",160, 220);
  }
  
  void Start() {
   if (_state == state_init) {
     Run();    
     return;
   }
  }
  
  void Move (int k) {
    if (_state == state_run) {
      if (k == 'q') _maze.goWest();
      else if (k == 's') _maze.goEast();
      else if (k == 'w') _maze.goSouth();
      else if (k == 'z') _maze.goNorth();
    }
  }   
 
  void Run() {
    _state = state_run;
    _startTime = millis();
    
    // Clear and draw score
    ClearTextArea();
  }
  
  void End() {
    _state = state_end;
    _endTime = millis();
    
    ClearTextArea();
    
    textAlign(CENTER);
    fill(black);
    text("FINISHED in",160, 180);
    int delta = (_endTime - _startTime) / 1000;
    int m = delta / 60;
    int s = (delta - m*60);
    String ti = "Time : " + m + "'" + s + "\"";
    text(ti ,160, 200);
    String p = "Current : " + _maze.getStep() + " steps";
    text(p ,160, 220);
    String d = "Best : " + _maze.getMaxDistance () + " steps";
    text(d ,160, 240);
  }
  
  void ClearTextArea () {
    fill (white);
    rect(0, 160, 320, 60);
  }
  
  void KeyPressed (int k) {
    if (k == 'r') Reset(); // Resetting game
    if (k == ' ') Start(); // Start
    
    Move (k);  
  }
  
  void Manage() {
    if (_state == state_run) {
      if (_maze.AtEnd()) End();
      else { // Updating current time
        fill (white);
        rect(0, 180, 320, 20);
        fill (black);
        int delta = (millis() - _startTime) / 1000;
        int m = delta / 60;
        int s = (delta - m*60);
        String ti = "Time : " + m + "'" + s + "\"";
        text(ti ,160, 200);        
      }
    }
  }
  
  int _state;
  boolean _needToRedraw;
  
  int _startTime;
  int _endTime;
  
  Maze _maze;
};

int VIDE=0;
int MUR=1;
int PAS=2;
//=============== NODE ================
class Node {
  int _x;
  int _y;
  int _dir;
  int _distance;
  
  Node (int x, int y, int dir, int distance)  { 
    _x = x;
    _y = y;
    _dir = dir;
    _distance = distance;
  }

  int getX () { return _x; }
  int getY () { return _y; }
  int getDir () { return _dir; }
  int getDistance () { return _distance; }
  
};

//=============== MAZE ================
class Maze {
  Maze (int h, int w, int p) {
    _h = h;
    _w = w;
    _sx = 1;
    _sy = 1;
    _dirs = 0;
    _p = p;
    
    if (_w < 5) _w = 5;
    if (_h < 5) _h = 5;

    _m = new int [_h][_w];
    _nodes = new ArrayList();
    
    reset();
  }

  
void show (int dimension) {
 
  if (dimension > 1 && dimension < 10) _d = dimension;
  
  // Maze
  for (int j = 0; j < _h; ++j) {
    for (int k = 0; k < _w; ++k) {
      color col = red;
      int val = _m[j][k];
      if (val == VIDE || val == MUR) col = color (0x80, 0x80, 0x80);
      else if (val == PAS) col = color (0xFF, 0xFF, 0xFF);
      
      rectMode(CORNER); 
      fill(col);
      rect(k*_d, j*_d, _d, _d);
    }
  }
  
  // Starting point
  fill(red);
  rect(_sx*_d, _sy*_d, _d, _d);
  
  // Ending point
  fill(green);
  rect(_ex*_d, _ey*_d, _d, _d);
  }

  void checkNode (Node aNode) {
      int x = aNode.getX();
      int y = aNode.getY();
      int distance = aNode.getDistance();
      
      switch (aNode.getDir()) {
      case 1: if (TestN (x, y-1)) addNode (x, y-1, distance+1); break; // N
      case 2: if (TestE (x+1, y)) addNode (x+1, y, distance+1); break; // E
      case 3: if (TestS (x, y+1)) addNode (x, y+1, distance+1); break; // S
      case 4: if (TestW (x-1, y)) addNode (x-1, y, distance+1); break; // W
      default: break;
      }
  }
  
  boolean TestN (int x, int y) {
    if (_m[y][x] != VIDE) return false;
    if (_m[y-1][x] == PAS) return false;
    if (_m[y][x-1] == PAS) return false;
    if (_m[y][x+1] == PAS) return false;
    return true;
  }

  boolean TestE (int x, int y) {
    if (_m[y][x] != VIDE) return false;
    if (_m[y][x+1] == PAS) return false;
    if (_m[y-1][x] == PAS) return false;
    if (_m[y+1][x] == PAS) return false;
    return true;
}
  
  boolean TestS (int x, int y) {
    if (_m[y][x] != VIDE) return false;
    if (_m[y][x-1] == PAS) return false;
    if (_m[y][x+1] == PAS) return false;
    if (_m[y+1][x] == PAS) return false;
    return true;
  }
  
  boolean TestW (int x, int y){
    if (_m[y][x] != VIDE) return false;
    if (_m[y][x-1] == PAS) return false;
    if (_m[y-1][x] == PAS) return false;
    if (_m[y+1][x] == PAS) return false;
    return true;
  }

  void reset () {
    for (int y = 0; y < _h; ++y) {
      for (int x = 0; x < _w; ++x) {
        _m[y][x]= VIDE;
      }
    }

    for (int y= 0; y < _h; ++y) {
      _m[y][0] = MUR;
      _m[y][_w-1] = MUR;
    }

    for (int x= 0; x < _w; ++x) {
      _m[0][x] = MUR;
      _m[_h-1][x] = MUR;
    }

    _sx = int (random (1, _w-1));
    _sy = int (random (1, _h-1));
    _ex = _sx;
    _ey = _sy;
    _mx = _sx;
    _my = _sy;
    
    _maxdistance=0;
    
     _d = 4;
    _dirs = int (random (0, 24));
  }

  void addNode (int x, int y, int distance) {
    // Compute new directions
    if (_p > 1) {
      if (int(random (0, _p)) == 0) { 
  _dirs = int(random (0, 24)); // Select moves order 
      }
    }
    else {
      _dirs = int (random (0, 24)); // Select moves order 
    }
    
    for (int idx = 0; idx < 4; ++idx) {
      int d = dirset[_dirs][idx];
      _nodes.add (new Node (x, y, d, distance));  // Adds 4 Nodes 
    }

    _m[y][x] = PAS; // OK we walked on it
    
    if (distance > _maxdistance) {
      _maxdistance = distance;
      _ex = x;
      _ey = y;
    }
  }

  void compute () {
    reset ();
    
    addNode (_sx, _sy, 0); // inserting first node
    while (true) {
      int n = _nodes.size();
      if (n == 0) return; // The end
      
      Node node = (Node) _nodes.get(n - 1); // Taking last one
      _nodes.remove (n - 1);
      checkNode (node);
    }
  }
  
  int getCell (int y, int x) {
    if (x >= _w || x < 0) return 0;
    if (y >= _h || y < 0) return 0;
    return _m[y][x];
  }
  
  int getMaxDistance () { return _maxdistance; }
  int getStep () { return _step; }
  
  boolean AtEnd() {
    if (_my != _ey) return false;
    if (_mx != _ex) return false;
    return true;
  }
  
  void goWest () {
     if (_m[_my][_mx-1]==PAS) {
        _step++;
       fill(white);
       rect(_mx*_d, _my*_d, _d, _d);
       _mx--;
       fill(blue);
       rect(_mx*_d, _my*_d, _d, _d);
     }
  } 
  
       void goEast () {
         if (_m[_my][_mx+1]==PAS) {
            _step++;
           fill(white);
           rect(_mx*_d, _my*_d, _d, _d);
           _mx++;
           fill(blue);
           rect(_mx*_d, _my*_d, _d, _d);
         }
      } 

  void goNorth () {
     if (_m[_my-1][_mx]==PAS) {
        _step++;
       fill(white);
       rect(_mx*_d, _my*_d, _d, _d);
       _my--;
       fill(blue);
       rect(_mx*_d, _my*_d, _d, _d);
     }
  } 
  
       void goSouth () {
         if (_m[_my+1][_mx]==PAS) {
            _step++;
           fill(white);
           rect(_mx*_d, _my*_d, _d, _d);
           _my++;
           fill(blue);
           rect(_mx*_d, _my*_d, _d, _d);
         }
      } 
      
  int [][]_m;
  int _h, _w; // H & W
  int _sx, _sy; // Starting point
  int _ex, _ey; // Ending point
  
  int _maxdistance; // Max distance between starting and ending point
  
  int _d; // Drawing size for cells  
  
  // Navigation
  int _step; // user steps
  int _mx, _my; // Current user position
  
  int _dirs;
  int _p; // Change direction probablility : 1->each step, 4-> 1/4 step
  ArrayList _nodes;
};

int [][] dirset = {
    { 1, 2, 3, 4},
    { 1, 2, 4, 3},
    { 1, 3, 2, 4},
    { 1, 3, 4, 2},
    { 1, 4, 2, 3},
    { 1, 4, 3, 2},
    
    { 2, 1, 3, 4},
    { 2, 1, 4, 3},
    { 2, 3, 1, 4},
    { 2, 3, 4, 1},
    { 2, 4, 1, 3},
    { 2, 4, 3, 1},

    { 3, 1, 2, 4},
    { 3, 1, 4, 2},
    { 3, 2, 1, 4},
    { 3, 2, 4, 1},
    { 3, 4, 1, 2},
    { 3, 4, 2, 1},
    
    { 4, 1, 2, 3},
    { 4, 1, 3, 2},
    { 4, 2, 1 ,3},
    { 4, 2, 3, 1},
    { 4, 3, 1, 2},
    { 4, 3, 2, 1}
};

//=============== MAIN ================
Game game;
PFont font;
void setup () {
  size (320,240);
 
  colorMode(RGB, height, height, height);
  background(white);
  noFill();
  noStroke ();
  
  font = createFont("Arial",16,true);  // Loading font
  textFont(font);
  
  game = new Game ();
}

void draw () { 
  game.Manage();
}

void keyPressed() {
  game.KeyPressed (key);
}
